<?php

namespace Drupal\mcapi_forms;
use Drupal\mcapi_forms\Entity\McForm;
use Drupal\mcapi\Form\TransactionForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\mcapi\Plugin\Field\FieldWidget\EntryWidget;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Entity\EntityConstraintViolationListInterface;

/**
 * Form builder for transaction formed designed via the UI.
 */
class DesignedTransactionForm extends TransactionForm {
  protected $config;

  /**
   * {@inheritDoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\Core\Entity\Entity\EntityFormDisplay $display */
    $display = $form_state->get('form_display');
    $form['#title'] = $this->config->label();
    $form['#twig_template'] = str_replace(["\r\n","\n","\r"], "<br/>", $this->config->getTwigTemplate());
    // avoid calling parent::form and adding the fields.
    $form['#process'][] = '::processForm';
    $form['#entity_builders']['update_form_langcode'] = '::updateFormLangcode';
    $form['#parents'] = [];
    \Drupal::service('renderer')->addCacheableDependency($form, $display);

    // Place entry fields (payer, payee, description, quant) at the top level of the form.
    foreach (EntryWidget::rawElements($this->entity->entries, 0) as $id => $element)  {
      $tok = substr($id, -4) == '_wid' ? substr($id, 0, -4) : $id;
      if (preg_match('/\{\{ ?'.$tok.' ?\}\}/', $form['#twig_template'])) {
        $form['#tokens'][$id] = $tok;
        // this wrapper is needed for css to highlight the errors.
        $element['#prefix'] = '<span class="form-item">';
        $element['#suffix'] = '</span>';
        $form[$id] = $element;
      }
    }

    // Place field API fields (No base fields can appear on Designed forms.)
    $added_fields = array_filter(
      \Drupal::service('entity_field.manager')->getFieldDefinitions('mc_transaction', 'mc_transaction'),
      function ($f) {return !$f instanceOf \Drupal\Core\Field\BaseFieldDefinition;}
    );
    foreach ($added_fields as $field_definition) {
      $fname = $field_definition->getName();
      if (preg_match('/\{\{ ?'.$fname.' ?\}\}/', $form['#twig_template'])) {
        $form['#tokens'][$fname] = $fname;
        $items = $this->getEntity()->get($fname);
        $element = ['#title_display' => 'hidden', '#required' => TRUE];// Assume all fields are required.
        $form[$fname] = $display->getRenderer($fname)
          ->formElement($items, 0, $element, $form, $form_state);
        // remove the container i.e. the field label.
        if (isset($form[$fname]['widget']['0']['value'])) {
          $form[$fname] = $form[$fname]['widget']['0']['value'];
        }
        if (isset($form[$fname]['widget']['0']['target_id'])) {
          $form[$fname] = $form[$fname]['widget']['0']['target_id'];
        }
        elseif (isset($form[$fname]['widget'])) {
          $form[$fname] = $form[$fname]['widget'][0];
        }
      }
      else $form[$fname]['#access'] = FALSE;
    }
    $form['#twig_template'] = str_replace(["\r\n","\n","\r"], "<br/>", $this->config->getTwigTemplate());
    $form['#attached']['library'][] = 'mcapi_forms/display';
    $form['#after_build'][] = '::prepareTemplateForm';
    // Links are not being detected in the yml file for some reason.
    $form['#contextual_links']['mcapi_form'] = ['route_parameters' => []];
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = parent::actions($form, $form_state);
    $actions['submit']['#value'] = $this->t('Send');
    // @todo check this is working.
    $form['#cache']['contexts'][] = 'user';
    return $actions;
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'designed_transaction_form';
  }


  /**
   * Provide an ID for the transaction form consistent with the keys of hook_transction_form_list().
   */
  function transactionFormId() {
    return 'designed_'.$this->config->id();
  }

  /**
   * {@inheritDoc}
   */
  public function buildEntity(array $form, FormStateInterface $form_state) {
    $form_state->setValue('entries', [
      'description' => $form_state->getValue('description'),
      'quant' => $form_state->getValue('quant'),
      'payee_wid' => $form_state->getValue('payee_wid'),
      'payer_wid' => $form_state->getValue('payer_wid'),
      'primary' => TRUE
    ]);
    return parent::buildEntity($form, $form_state);
  }

  /**
   * Form #afterbuild callback
   * Break up the template into render elements alternating between markup and form elements.
   */
  function prepareTemplateForm($form, $form_state) {
    // move the form elements to be placed from tokens out of the form and into
    // $replacements, stripping off the containers.
    foreach ($form['#tokens'] as $fname => $tok) {
      $replacements[$tok] = $form[$fname];
      unset($form[$fname]);
    }

    $w = 0;
    // break up the template into alternating tokens and html snippents
    $tok = strtok(' '.$form['#twig_template'], '{');// ensure the first is text
    while ($tok !== FALSE) {
      $parts[] = $tok;
      $tok = strtok('{}');
    }
    foreach ($parts as $key => $snippet) {
      $snippet = trim($snippet);
      if ($key % 2 == 0) { //even numbers are text
        $form['snippet-'.$key] = [
          '#markup' => $snippet,
          '#weight' => $w++,
          '#parents' => []
        ];
      }
      // Replace the fields
      elseif (isset($replacements[$snippet])) { // odd numbers are tokens
        $element = $replacements[$snippet];
        $element['#weight'] = $w++;
        if (empty($element['#parents'])) {
          $element['#parents'] = $element['#array_parents'] = [$element['#name']];
        }
        $element['#array_parents']= $element['#parents'];
        unset($element['#theme_wrappers']);
        $form[$snippet] = $element;
      }
      else {
        throw new \Exception("Unrecognised token in designed transaction form: $snippet");
      }
    }
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function getEntityFromRouteMatch(RouteMatchInterface $route_match, $entity_type_id) {
    $route_name = $route_match->getRouteName();
    $mc_form_id = substr($route_name, strrpos($route_name, '.') + 1);
    $this->config = McForm::load($mc_form_id);
    $this->request->attributes->set('mc_workflow', $this->config->getWorkflowId());
    return parent::getEntityFromRouteMatch($route_match, $entity_type_id);
  }

  /**
   * Ensure that errors to the payer, payee, quant and description fields are processed.
   */
  protected function getEditedFieldNames(FormStateInterface $form_state) {
    $fields = parent::getEditedFieldNames($form_state);
    $fields[] = 'entries';
    return $fields;
  }

  /**
   * {@inheritDoc}
   */
  protected function flagViolations(EntityConstraintViolationListInterface $violations, array $form, FormStateInterface $form_state) {
    foreach ($violations as $offset => $violation) {
      $propPath = $violation->getPropertyPath();
      $fieldname = substr($propPath, strrpos($propPath, '.') + 1);
      switch ($fieldname) {
        // Drupals' form.css expects these fields to be in a <div class="form-item"> to show the error message
        case 'payee_wid':
          $element = $form['payee'];
          break;
        case 'payer_wid':
          $element = $form['payer'];
          break;
        case 'quant':
        case 'description':
          $element = $form[$fieldname];
          break;
        default:
          break 2;
      }
      $form_state->setError($element, $violation->getMessage());
      $violations->remove($offset);
    }
    parent::flagViolations($violations, $form, $form_state);

  }
}
